/*
 * DamlModel.java
 *
 * Created on February 17, 2002, 3:10 PM
 */

package org.semanticweb.daml;

import java.util.*;
import java.io.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.jena.daml.*;
import com.hp.hpl.jena.daml.common.*;
import org.semanticweb.dl.*;


/**
 * Wrapper for Jena's DAMLModel.
 * @author  Michael Sintek
 * @version 
 */
public class DamlModel {

  DAMLModel _model;
  KB _kb;
  
  Symbol TOP = Symbol.TOP;
  Symbol BOTTOM = Symbol.BOTTOM;
  Symbol UNKNOWN = Symbol.UNKNOWN;

  static final String DLINPUT = "#dlinput.xml"; // input file for DL classifier
  static final String DLTMP = "#dloutput.xml"; // temp file for DL classifier
  
  
  /** Creates new DamlModel */
  public DamlModel(java.io.Reader in, java.lang.String base) throws Exception {
    _model = new DAMLModelImpl();
    _model.read(in, base);
    _model.setUseEquivalence(false); // JenaBUG: does not (always?) work!!
  }

  
  public KB toDL() { // transformation to DL (package org.semanticweb.dl)
    _kb = new KB();
    if (_model != null) {
      handleClasses();
      handleProperties();
      // handle instances when they are supported by DL engine
      /* ???
      for (Iterator instIter = _model.listDAMLInstances(); instIter.hasNext();) {
        Object inst = instIter.next();
        System.err.println("instance ignored: " + inst);
      }
      */
    }
    return _kb;
  }
  
  
  // class handling
  
  void handleClasses() {
    for (Iterator classIter = _model.listDAMLClasses(); classIter.hasNext();) {
      DAMLClass damlClass = (DAMLClass)classIter.next();
      if (damlClass.isNamedClass()) {
        Symbol classSymbol = _kb.getSymbol(damlClass);
        if (!classSymbol.isTop() && !classSymbol.isBottom()) {
          _kb.add(new Defconcept(classSymbol));
          _kb.addConcept(classSymbol);
        }
        // as I understand the DAML+OIL definition, anonymous classes
        // are not allowed here ... ???
        handleSuperClasses(damlClass);
        handleEquivalentClasses(damlClass);
        handleDisjointWith(damlClass);
        // union, disjointUnion, intersection, complement
        // -> (equal c (functor ....)) assertions
	ArrayList list = new ArrayList();
        addUnionOfList(damlClass, list);
        addOneOfList(damlClass, list);
        addDisjointUnionOfList(damlClass, list);
        addIntersectionOfList(damlClass, list);
        addComplementOfList(damlClass, list);
        for (Iterator iter = list.iterator(); iter.hasNext();) {
          _kb.add(new Equal(classSymbol, (Expression)iter.next()));
        }
      }
    }
  }
  
  void handleSuperClasses(DAMLClass damlClass) {
    Expression classExpression = getExpression(damlClass);
    for (Iterator superIter = damlClass.getSuperClasses(false); superIter.hasNext();) {
      RDFNode superClass = (RDFNode)superIter.next();
      _kb.add(new Implies(classExpression, getExpression(superClass)));
    }
  }
  
  void handleEquivalentClasses(DAMLClass damlClass) {
    Expression classExpression = getExpression(damlClass);
    for (Iterator equivIter = damlClass.getEquivalentValues(); equivIter.hasNext();) {
      RDFNode equiv = (RDFNode)equivIter.next();
      Expression equivExpression = getExpression(equiv);
      if (equivExpression instanceof Symbol) { 
        // only one "direction" is added to KB
        if (((Symbol)classExpression).getFullName().compareTo(((Symbol)equivExpression).getFullName()) < 0)
	  _kb.add(new Equal(classExpression, getExpression(equiv)));
      } else
        _kb.add(new Equal(classExpression, getExpression(equiv)));
    }
  }
  
  void handleDisjointWith(DAMLClass damlClass) {
    Expression classExpression = getExpression(damlClass);
    ArrayList list = getList(damlClass.prop_disjointWith());
    for (Iterator iter = list.iterator(); iter.hasNext();)
      // cannot be used in an Equal ???
      _kb.add(new Disjoint(classExpression, (Expression)iter.next()));
  }
  
  void addDisjointUnionOfList(DAMLClass damlClass, ArrayList l) {
    ArrayList list = getList(damlClass.prop_disjointUnionOf());
    for (Iterator iter = list.iterator(); iter.hasNext();) {
      Expression duList = (Expression)iter.next();
      if (duList instanceof ExpressionList)
        l.add(new DisjointUnion(((ExpressionList)duList).getList()));
      else
        System.err.println("list expected: " + duList);
    }
  }
  
  void addUnionOfList(DAMLClass damlClass, ArrayList l) {
    ArrayList list = getList(damlClass.prop_unionOf());
    for (Iterator iter = list.iterator(); iter.hasNext();) {
      Expression uList = (Expression)iter.next();
      if (uList instanceof ExpressionList)
        l.add(Or.getOr(((ExpressionList)uList).getList()));
      else
        System.err.println("list expected: " + uList);
    }
  }

  void addOneOfList(DAMLClass damlClass, ArrayList l) {
    // handle like unionOf at the moment
    ArrayList list = getList(damlClass.prop_oneOf());
    for (Iterator iter = list.iterator(); iter.hasNext();) {
      Expression uList = (Expression)iter.next();
      if (uList instanceof ExpressionList) {
	System.err.println("oneOf approximated as disjunction: " + uList);
        l.add(Or.getOr(((ExpressionList)uList).getList(), BOTTOM)); // BOTTOM ???
      } else
        System.err.println("list expected: " + uList);
    }
  }
  
  void addIntersectionOfList(DAMLClass damlClass, ArrayList l) {
    ArrayList list = getList(damlClass.prop_intersectionOf());
    for (Iterator iter = list.iterator(); iter.hasNext();) {
      Expression iList = (Expression)iter.next();
      if (iList instanceof ExpressionList)
        l.add(And.getAnd(((ExpressionList)iList).getList()));
      else
        System.err.println("list expected: " + iList);
    }
  }
  
  void addComplementOfList(DAMLClass damlClass, ArrayList l) {
    ArrayList list = getList(damlClass.prop_complementOf());
    for (Iterator iter = list.iterator(); iter.hasNext();)
      l.add(new Not((Expression)iter.next()));
  }


  // property handling
  
  void handleProperties() {
    for (Iterator propIter = _model.listDAMLProperties(); propIter.hasNext();) {
      DAMLProperty dprop = (DAMLProperty)propIter.next();
      Expression property = getExpression(dprop);
      // System.err.println("property: " + property + " = " + dprop);
      if (property instanceof Symbol) { 
        // not for property expressions ???
        _kb.add(new Defrole((Symbol)property));
        _kb.addRole((Symbol)property);
	// handle domain
	ArrayList domain = getList(dprop.prop_domain());
	if (!domain.isEmpty())
	  _kb.add(new Domain(property, And.getAnd(domain)));
	// handle super properties
	for (Iterator superIter = dprop.getSuperProperties(false); superIter.hasNext();) {
	  RDFNode superProperty = (RDFNode)superIter.next();
	  _kb.add(new ImpliesR(property, getExpression(superProperty)));
	}
	// handle equivalent properties
	for (Iterator equivIter = dprop.getEquivalentValues(); equivIter.hasNext();) {
	  RDFNode equiv = (RDFNode)equivIter.next();
	  Expression equivExpression = getExpression(equiv);
	  if (equivExpression instanceof Symbol) {
	    // only one "direction"
	    if (((Symbol)property).getFullName().compareTo(((Symbol)equivExpression).getFullName()) < 0)
	      _kb.add(new EqualR(property, equivExpression));
	  } else
	    _kb.add(new EqualR(property, equivExpression));
	}
	// handle unique
	if (dprop.isUnique())
	  _kb.add(new Unique(property));        
	// -- for ObjectProperties:
	if (dprop instanceof DAMLObjectProperty) {
	  DAMLObjectProperty oprop = (DAMLObjectProperty)dprop;
	  // handle range
	  ArrayList range = getList(oprop.prop_range());
	  if (!range.isEmpty())
	    _kb.add(new Range(property, And.getAnd(range)));
	  // handle transitive
	  if (oprop.isTransitive())
	    _kb.add(new Transitive(property)); 
	  // handle unambiguous
	  if (oprop.isUnambiguous()) 
	    _kb.add(new Unambiguous(property));
	  // handle inverse
	  ArrayList inverses = getList(oprop.prop_inverseOf());
	  if (!inverses.isEmpty()) {
	    for (Iterator iter = inverses.iterator(); iter.hasNext();) {
	      Expression inverse = (Expression)iter.next();
	      // (equalr r1 (inv r2))
	      _kb.add(new EqualR(property, new Inverse(inverse)));
	    }
	  }
	} else { // != ObjectProperty -> Datatype Property
	  // if (dprop instanceof DAMLDatatypeProperty) // (not a :-) JenaBUG
	  // but we simply assume that all other proerties are datatype properties!
	  // System.err.println("datatype properties not correctly supported: " 
	  //   + property); // + " (" + dprop + ")");
	  // we simply make the range a normal class (which is not correct):
	  ArrayList range = getList(dprop.prop_range());
	  if (!range.isEmpty())
	    _kb.add(new Range(property, And.getAnd(range)));
	}
      }
    }
  }
  
  
  // general expression handling
  
  Expression getExpression(RDFNode obj) {
    if (obj instanceof DAMLClassExpression) {
      DAMLClassExpression classExpr = (DAMLClassExpression)obj;
      if (classExpr.isNamedClass()) {
        return _kb.getSymbol((Resource)classExpr);
      } else { // anonymous class expression
        // this can be more than one expression, so we need a set (-> And)
        ArrayList set = new ArrayList();
        // go through all possibilities form DAMLClassExpression:
        // isComplement
        if (classExpr.isComplement())
          addComplementOfList((DAMLClass)classExpr, set);
        // isDisjointUnion
        if (classExpr.isDisjointUnion())
          addDisjointUnionOfList((DAMLClass)classExpr, set);
        // isEnumeration (= oneOf):
        if (classExpr.isEnumeration()) {
          addOneOfList((DAMLClass)classExpr, set);
        }
        // isIntersection
        if (classExpr.isIntersection())
          addIntersectionOfList((DAMLClass)classExpr, set);
        // is Restriction
        if (classExpr.isRestriction())
          addRestrictions((DAMLRestriction)classExpr, set);
        // isUnion
        if (classExpr.isUnion())
          addUnionOfList((DAMLClass)classExpr, set);
        return And.getAnd(set, TOP); // no restriction -> TOP
      }
    } else if (obj instanceof DAMLProperty) {
      if (((DAMLProperty)obj).isAnon()) {
        if (obj instanceof DAMLObjectProperty) {
	  Expression inverse = getSingleton(((DAMLObjectProperty)obj).prop_inverseOf());
	  return new Inverse(inverse);
	} else // what other anonymous property expressions are allowed? domain/range?
	  return UNKNOWN;
      } else
	return _kb.getSymbol((Resource)obj);
    } else if (obj instanceof DAMLList) {
      return getExpressionList((DAMLList)obj);
    // else ... ???
    } else if (obj instanceof Resource) { // normal symbol
      return _kb.getSymbol((Resource)obj);
    } else { // Literal ... ????
      System.err.println("illegal expression: " + obj);
      return UNKNOWN;
    }
  }

  void addRestrictions(DAMLRestriction restriction, ArrayList set) {
    PropertyAccessor onPropertyAccessor = restriction.prop_onProperty();
    Expression onProperty = getSingleton(onPropertyAccessor);
    // more than one of the following cases is allowed
    // toClass -> All
    PropertyAccessor accessor = restriction.prop_toClass();
    if (accessor.count() > 0)
      set.add(new All(onProperty, And.getAnd(getList(accessor))));
    // hasValue -> not yet correctly supported
    // we map to (some p v) for each value
    accessor = restriction.prop_hasValue();
    if (accessor.count() > 0) {
      // set.add(TOP); // not necessary: empty restriction = TOP
      List values = getList(accessor);
      for (Iterator iter = values.iterator(); iter.hasNext();) {
        Expression value = (Expression)iter.next();
	System.err.println("hasValue approximated as \"some\": " + value);
	set.add(new Some(onProperty, value));
      }
    }
    // hasClass -> Some
    accessor = restriction.prop_hasClass();
    if (accessor.count() > 0)
      set.add(new Some(onProperty, And.getAnd(getList(accessor))));
    // minCardinality -> AtLeast
    IntLiteralAccessor intAccessor = restriction.prop_minCardinality();
    if (intAccessor.count() > 0) {
      int card = intAccessor.getInt();
      set.add(new AtLeast(card, onProperty));
    }
    // maxCardinality -> Atmost
    intAccessor = restriction.prop_maxCardinality();
    if (intAccessor.count() > 0) {
      int card = intAccessor.getInt();
      set.add(new AtMost(card, onProperty));
    }
    // cardinality -> Exactly
    intAccessor = restriction.prop_cardinality();
    if (intAccessor.count() > 0) {
      int card = intAccessor.getInt();
      set.add(new Exactly(card, onProperty));
    }
    // minCardinalityQ & hasClassQ -> AtLeast
    intAccessor = restriction.prop_minCardinalityQ();
    if (intAccessor.count() > 0) {
      accessor = restriction.prop_hasClassQ();
      int card = intAccessor.getInt();
      // TOP: if hasClassQ is missing; should not happen!
      set.add(new AtLeast(card, onProperty, And.getAnd(getList(accessor), TOP)));
    }
    // maxCardinalityQ & hasClassQ -> Atmost
    intAccessor = restriction.prop_maxCardinalityQ();
    if (intAccessor.count() > 0) {
      accessor = restriction.prop_hasClassQ();
      int card = intAccessor.getInt();
      set.add(new AtMost(card, onProperty, And.getAnd(getList(accessor), TOP)));
    }
    // cardinalityQ & hasClassQ -> Exactly
    intAccessor = restriction.prop_cardinalityQ();
    if (intAccessor.count() > 0) {
      accessor = restriction.prop_hasClassQ();
      int card = intAccessor.getInt();
      set.add(new Exactly(card, onProperty, And.getAnd(getList(accessor), TOP)));
    }
  }
  
  
  // handling of daml lists
  Expression getExpressionList(DAMLList list) {
    ArrayList set = new ArrayList();
    for (Iterator iterator = list.getAll(); iterator.hasNext();) {
      RDFNode expr = (RDFNode)iterator.next();
      set.add(getExpression(expr));
    }
    return new ExpressionList(set);
  }
  
  
  ArrayList getList(PropertyAccessor propertyAccessor) {
    ArrayList list = new ArrayList();
    for (Iterator iterator = propertyAccessor.getAll(false); iterator.hasNext();) {
      RDFNode expr = (RDFNode)iterator.next();
      list.add(getExpression(expr));
    }
    /* JenaBUG: equals are returned!!
    if (list.size() > 1)
      System.err.println(propertyAccessor + " -> " + list);
    */
    return list;
  }
  
  Expression getSingleton(PropertyAccessor propertyAccessor) {
    int count = propertyAccessor.count();
    if (count != 1)
      System.err.println("only one value exprected: " + propertyAccessor);
    if (count > 0) {
      // System.out.println(propertyAccessor.getAll(false).next());
      // return getExpression(propertyAccessor.get()); // JenaBUG -> HACK:
      return getExpression((RDFNode)propertyAccessor.getAll(false).next());
    } else
      return UNKNOWN;
  }
  

  public static void main(String[] args) {
    if (args.length == 0) {
      System.err.println("arguments expected: [-xml|-lisp|-dot|-triple] [-racer <racer-pgm>]");
      System.err.println("                    (<filename>|-) [-ns <namespace>] [-o <outputfile>]");
      System.err.println("  -xml         : output in XML syntax according to FaCT DTD (default)");
      System.err.println("  -lisp        : output in LISP syntax");
      System.err.println("  -daml        : output in DAML syntax (only with -racer)");
      System.err.println("  -dot         : output as Graphviz dot file (only with -racer)");
      System.err.println("  -triple      : output in internal TRIPLE syntax (only with -racer)");
      System.err.println("  -racer <pgm> : send to racer for classification; returns taxonomy");
      System.err.println("  -            : read from standard input");
      System.err.println("  -ns <ns>     : use this as default namespace");
      System.err.println("  -o <out>     : write output to file (default: standard output)");
      System.exit(1);
    }
    boolean xml = true;
    boolean useracer = false;
    boolean triple = false;
    boolean dot = false;
    boolean daml = false;
    String racerpgm = "racer";
    String filename = null; // = read from standard input
    String outputFilename = null; // = write to standard output
    String namespace = "http://www.semanticweb.org/example#";
    int pos = 0;
    if (args[pos].equalsIgnoreCase("-xml")) {
      xml = true;
      pos++;
    } else if (args[pos].equalsIgnoreCase("-lisp")) {
      xml = false;
      pos++;
    } else if (args[pos].equalsIgnoreCase("-triple")) {
      xml = false;
      triple = true;
      pos++;
    } else if (args[pos].equalsIgnoreCase("-dot")) {
      xml = false;
      dot = true;
      pos++;
    } else if (args[pos].equalsIgnoreCase("-daml")) {
      xml = false;
      daml = true;
      pos++;
    }
    if (args.length >= pos+2) {
      if (args[pos].equalsIgnoreCase("-racer")) {
        useracer = true;
        racerpgm = args[pos+1];
        pos = pos+2;
      }
    }
    if (args.length >= pos+1) {
      filename = args[pos];
      if (filename.equals("-")) filename = null;
      pos++;
    }
    if (args.length >= pos+2) {
      if (args[pos].equalsIgnoreCase("-ns")) {
        namespace = args[pos+1];
        pos = pos+2;
      }
    }
    if (args.length >= pos+2) {
      if (args[pos].equalsIgnoreCase("-o")) {
        outputFilename = args[pos+1];
        pos = pos+2;
      }
    }
    try {
      System.err.println("*** Parsing input ...");
      Reader in;
      if (filename != null)
        in = new FileReader(filename);
      else
        in = new InputStreamReader(System.in);
      DamlModel model = new DamlModel(in, namespace);
      KB kb = model.toDL();
      // kb.print(); // for intermediate representation
      System.err.println("*** done.");
      System.err.println("\n*** Transforming to SHIQ ...");
      KB shiq = kb.toSHIQ();
      System.err.println("*** done.");
      PrintWriter out = null;
      if (!useracer && outputFilename == null)
        out = new PrintWriter(System.out);
      else {
        String ofn = null;
        if (useracer) ofn = DLINPUT;
        else          ofn = outputFilename;
        out = new PrintWriter(new FileWriter(ofn));
      }
      Map abbrevs = null;
      if (xml || useracer)
        abbrevs = shiq.printXML(out);
      else // we do not allow triple syntax here yet, so it must be LISP:
        abbrevs = shiq.printLISP(out);
      if (useracer) {
        System.err.println("\n*** Running RACER ...");
        // revert abbrevs
        HashMap namespaces = new HashMap();
        for (Iterator iter = abbrevs.entrySet().iterator(); iter.hasNext();) {
          Map.Entry entry = (Map.Entry)iter.next();
          namespaces.put(entry.getValue(), entry.getKey());
        }
        Racer racer = new Racer(racerpgm, DLINPUT, namespace, namespaces, DLTMP);
        racer.runRacer();
        System.err.println("*** done.");
        System.err.println("\n*** Analyzing RACER output ...");
        KB racerOutputKB = racer.getTaxonomyKB();
        System.err.println("*** done.");
        PrintWriter out2 = null;
        if (outputFilename == null)
          out2 = new PrintWriter(System.out);
        else
          out2 = new PrintWriter(new FileWriter(outputFilename));
        Syntax syntax;
        if (xml)         syntax = new XMLSyntax(racerOutputKB);
        else if (triple) syntax = new TripleSyntax(racerOutputKB);
        else if (dot)    syntax = new DotSyntax(racerOutputKB);
        else if (daml)   syntax = new DAMLSyntax(racerOutputKB);
        else             syntax = new Syntax(racerOutputKB);
        racerOutputKB.print(syntax, out2);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  
}

