/*
 * Racer.java
 *
 * Created on February 27, 2002, 11:31 AM
 */

package org.semanticweb.daml;

import java.util.*;
import java.io.*;
import org.w3c.dom.*;
import org.apache.xerces.parsers.DOMParser;
import org.xml.sax.InputSource;
import org.semanticweb.dl.*;


/**
 * Wrapper for DL classifier RACER. 
 * @author  Michael Sintek
 * @version 
 */
public class Racer {
  static final String QUERIESFILE = "#queries.racer";
  
  String _racerProgram;
  String _inputFileName;
  String _tmpFileName;
  String _namespace;
  Map _namespaces; // abbrev -> namespace (Strings)
  
  // for step 2: analyzing output
  Document _document;
  KB _kb;
  ArrayList _equals;
  
  
  /** Creates new Racer */
  public Racer(String racerProgram, String inputFileName, 
      String namespace, Map namespaces, String tmpFileName) {
    _racerProgram = racerProgram;
    _inputFileName = inputFileName;
    _tmpFileName = tmpFileName;
    _namespace = namespace;
    _namespaces = namespaces; 
  }
  
  // 1. run racer
  
  public void runRacer() throws Exception {
    PrintWriter writer = new PrintWriter(new FileWriter(QUERIESFILE));
    writer.println("(taxonomy)");
    writer.flush();
    writer.close();
    Process process =
      Runtime.getRuntime().exec(
        new String[] { _racerProgram,
                       "-xml", _inputFileName,
                       "-q", QUERIESFILE,
                       // "-t 10", // timeout in seconds
                       "-o", _tmpFileName } );
    process.getOutputStream().close(); // no user input
    // but we want to see the output:
    (new ShowOutputThread(process.getInputStream())).start();
    (new ShowOutputThread(process.getErrorStream())).start();
    int exitValue = process.waitFor();
    if (exitValue != 0)
      System.err.println("racer exited abnormally: " + exitValue);
  }
  
  
  // 2. analyze output -> KB
  
  public KB getTaxonomyKB() throws Exception {
    _document = parseXML(_tmpFileName);
    _kb = new KB();
    _equals = new ArrayList();
    Element element = _document.getDocumentElement();
    List children = getChildren(element);
    for (Iterator childIter = children.iterator(); childIter.hasNext();) {
      Element answer = (Element)childIter.next();
      String tagname = answer.getTagName();
      if (tagname.equals("ANSWER")) {
        String query = answer.getAttribute("QUERY");
        if (query.equalsIgnoreCase("(taxonomy)"))
          getTaxonomy(answer);
        else
          System.err.println("answer ignored: " + query);
      } else
        System.err.println("element not handled: " + tagname);
    }
    return _kb;
  }
  
  void getTaxonomy(Element answer) {
    // ANSWER element that contains result of (taxonomy) (or ERROR in next version)
    // (((concept conceptlist conceptlist) ...))
    // concept = symbol | (symbol1 symbol2 ...)
    // conceptlist = (concept1 concept2 ...)
    List answer1 = getChildren(answer);
    if (answer1.size() == 1) {
      List cds = getList((Element)answer1.iterator().next()); // concept descriptions
      for (Iterator cdIter = cds.iterator(); cdIter.hasNext();) {
        Element cd = (Element)cdIter.next();
        List conceptSuperSub = getList(cd);
        if (conceptSuperSub.size() == 3) {
          handleConceptDescription(
            (Element)conceptSuperSub.get(0),
            (Element)conceptSuperSub.get(1));
        } else {
          System.err.println("concept description list expected: " + cd);
        }
      }
    } else {
      System.err.println("answer ignored: " + answer);
    }
  }
  
  void handleConceptDescription(Element conceptElement, Element superConceptsElement) {
    List concept = getConceptList(conceptElement);
    List superConceptsList = getList(superConceptsElement);
    ArrayList superConcepts = new ArrayList();
    for (Iterator iter = superConceptsList.iterator(); iter.hasNext();)
      superConcepts.add(getConceptList((Element)iter.next()));
    // System.out.println(concept + " : " + superConcepts);
    // add to _kb
    addToKB(concept, superConcepts);
  }
  
  void addToKB(List concept, List superConcepts) {
    addConcept(concept);
    addEquals(concept);
    for (Iterator iter = superConcepts.iterator(); iter.hasNext();) {
      List superConcept = (List)iter.next();
      // addEquals(superConcept); // not necessary
      addSuper(concept, superConcept);
    }
  }

  
  // KB handling
  
  void addConcept(List concept) {
    for (Iterator iter = concept.iterator(); iter.hasNext();) {
      String conceptName = (String)iter.next();
      Symbol symbol = getSymbol(conceptName);
      // if (!symbol.isTop() && !symbol.isBottom())
      _kb.add(new Defconcept(getSymbol(conceptName)));
    }
  }
  
  void addEquals(List concept) {
    if (concept.size() > 1) { // not necessary: && !_equals.contains(concept)) {
      // _equals.add(concept); // not necessary
      Iterator iter = concept.iterator();
      String first = (String)iter.next();
      while (iter.hasNext()) {
        String other = (String)iter.next();
        addEqual(first, other);
      }
    }
  }
  
  void addSuper(List concept, List superConcept) {
    _kb.add(new Implies(getSymbol(concept), getSymbol(superConcept)));
  }
  
  void addEqual(String concept1, String concept2) {
    // System.out.println(concept1 + " = " + concept2);
    _kb.add(new Equal(getSymbol(concept1), getSymbol(concept2)));
  }
  
  Symbol getSymbol(String concept) {
    if (concept.equalsIgnoreCase("TOP"))
      return Symbol.TOP;
    else if (concept.equalsIgnoreCase("BOTTOM"))
      return Symbol.BOTTOM;
    else {
      // split ns+ln
      int pos = concept.indexOf('+');
      if (pos != -1) {
        String abbrev = concept.substring(0, pos);
        String namespace = (String)_namespaces.get(abbrev);
        if (namespace == null) // should never happen
          namespace = _namespace; // or: return new Symbol(_namespace, concept); ???
        String name = concept.substring(pos+1);
        return new Symbol(namespace, name);
      } else
        return new Symbol(_namespace, concept);
    }
  }
  
  Symbol getSymbol(List concept) {
    return getSymbol((String)concept.iterator().next()); // is never empty
  }
  
  
  // XML handling auxiliaries
  
  public Document parseXML(String fileName) throws Exception {
    InputStream in = new FileInputStream(_tmpFileName);
    DOMParser parser = new DOMParser();
    parser.parse(new InputSource(in));
    in.close();
    return parser.getDocument();
  }
  
  public List getChildren(Element element) {
    // gets all children which are elements
    ArrayList list = new ArrayList();
    NodeList children = element.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
      Node child = (Node)children.item(i);
      if (child instanceof Element)
        list.add(child);
    }
    return list;
  }
  
  List getList(Element listElement) {
    if (listElement.getTagName().equals("LIST"))
      return getChildren(listElement);
    else {
      System.err.println("list expected: " + listElement);
      return null;
    }
  }
      
  List getConceptList(Element concept) {
    // the result list is sorted alphabetically (so the first element
    // can easily be used as representative)
    // symbol | (symbol1 symbol2 ...)
    ArrayList list = new ArrayList();
    if (concept.getTagName().equals("SYMBOL"))
      list.add(getSymbolName(concept));
    else {
      List concepts = getList(concept);
      for (Iterator cIter = concepts.iterator(); cIter.hasNext();)
        list.add(getSymbolName((Element)cIter.next()));
    }
    Collections.sort(list);
    return list;
  }
  
  String getSymbolName(Element symbol) {
    if (symbol.getTagName().equals("SYMBOL"))
      return symbol.getAttribute("NAME");
    else {
      System.err.println("SYMBOL expected: " + symbol);
      return null;
    }
  } 
    
}


class ShowOutputThread extends Thread {
  // shows the standard or error output of a racer process
  
  InputStream _in;
  
  ShowOutputThread(InputStream in) {
    _in = in;
  }
  
  public void run() {
    try {
      int c = _in.read();
      while (c != -1) {
        System.err.print((char)c);
        c = _in.read();
      }
      _in.close();
    } catch (Exception e) {
      System.err.println(e);
    }
  }  
  
}

